/*
 The circuit:
  * SD card attached to SPI bus as follows:
 ** MOSI - pin 11
 ** MISO - pin 12
 ** CLK - pin 13 
 ** CS - pin 10 
*/
#include <SD.h>
#include "stk500.h"
#define HVDIR "HACKV"
#define LOADER "/HACKV/LOADER.HEX"

#define MAXFILES 16

const int chipSelect = 10;    
char names[MAXFILES][9];
char names_count = 0;
char page_num = 0;
int cnt = 0;
char filename[20];
bool loader_ready = false;

const unsigned int MAX_INPUT = 20;

// 57600 115200
#define BOOT_BAUD 57600
// different pins will be needed for I2SD, as 2/3 are leds
#define rstPin 5
//#define bootPin 4

// set up variables using the SD utility library functions:
// SdFile root;
File myFile;
avrmem mybuf;
unsigned char mempage[128];

void getDirectory(File dir) {
   while(true) {
     File entry =  dir.openNextFile();
     if (!entry || names_count>=MAXFILES) {
       break;
     }
     if (!entry.isDirectory()) {
       char *fname = entry.name();
       char fext[4];
       memcpy(&fext[0],fname + String(fname).indexOf('.') + 1,3);
       if (strcmp(fext,"HEX")==0) {
         if (strcmp(fname,"LOADER.HEX")!=0) {
         memcpy(&names[names_count++],fname,String(fname).indexOf('.'));
         }
       }
     }
   }
}



void setup()
{
  mybuf.buf = &mempage[0];
  Serial.begin(BOOT_BAUD);
  pinMode(rstPin,OUTPUT);
  pinMode(chipSelect,OUTPUT);
  SD.begin(10);
  myFile=SD.open(HVDIR);
  getDirectory(myFile);
}


// here to process incoming serial data after a terminator received
void process_data (char * data)
  {
    if ( strcmp(data,"list")==0 && names_count>0){
      delay(100);
      Serial.println(names_count,DEC);
      delay(100);
      for (int i=0;i<names_count;i++) {
        Serial.println(names[i]);
        delay(50);
      }
    }
    if ( strcmp(data,"progme")==0 ){
       Serial.flush();
       delay(200);
       programArduino(LOADER);
    }
    if ( strcmp(data,"ready")==0 && !loader_ready){
      loader_ready = true;
    }

    char f4[5];
    memcpy(&f4[0],data,4);
    f4[4]='\0';
    if (strcmp(f4,"load")==0) {
      memcpy(&f4[0],data+4,3);
       char fnum = atoi(f4);
       Serial.println(fnum,DEC);
       strcpy(filename,"/");
       strcat(filename,HVDIR);
       strcat(filename,"/"); 
       strcat(filename,names[fnum]);
       strcat(filename,".HEX");
       Serial.flush();
       delay(200);
       programArduino(filename);
    }
  }  // end of process_data
  

char pings = 0;
static char input_line [MAX_INPUT];
static unsigned int input_pos = 0;

void loop()
{
if (cnt++>1000) {  
  Serial.println("host ping!");
  cnt=0;
  pings++;
  if (pings==4 && !loader_ready) {
    Serial.flush();
    delay(200);
    programArduino(LOADER); 
  }
}


  if (Serial.available () > 0) 
    {
    char inByte = Serial.read ();

    switch (inByte)
      {

      case '\n':   // end of text
        input_line [input_pos] = 0;  // terminating null byte
        
        // terminator reached! process input_line here ...
        process_data (input_line);
        
        // reset buffer for next time
        input_pos = 0;  
        break;
  
      case '\r':   // discard carriage return
        break;
  
      default:
        // keep adding if not full ... allow for terminating null byte
        if (input_pos < (MAX_INPUT - 1))
          input_line [input_pos++] = inByte;
        break;

      }  

  }  
//  if (digitalRead(bootPin)==HIGH) {
//        Serial.flush();
//        delay(200);
//        programArduino(LOADER); 
//  }

  delay(1);
}  // end of loop

// *********************************************************************** 


// Line Buffer is set up in global SRAM
#define LINELENGTH 50
unsigned char linebuffer[LINELENGTH];
unsigned char linemembuffer[16];
int leftover = 0;
unsigned char leftoverbuf[16];
int readPage(File input, AVRMEM *buf)
{
  int len;
  int address;
  int total_len = leftover;
  if (leftover > 0){
    memcpy(buf->buf,leftoverbuf,leftover);
  }
  // grab 128 bytes or less (one page)
  int i = 0;
  while (total_len < 128){
    len = readIntelHexLine(input, &address, &linemembuffer[0]);
    if (len < 0){
      if (i==0) total_len = len;
      break;
    }
//    else
//      total_len=total_len+len;
    if (i==0)// first record determines the page address
    {
      buf->pageaddress = address - leftover;
      leftover = 0;
    }
    i++;
//    memcpy((buf->buf)+(i*16), linemembuffer, len);
    if ((total_len+len) > 128)
    {
      leftover = (total_len+len)-128;
      memcpy(buf->buf+total_len, linemembuffer,128-total_len);
      int kk = 0;
      for (int k =(128-total_len);k<len;k++)
        leftoverbuf[kk++] = linemembuffer[k];
      leftover = kk;  
      total_len = 128;
    }
    else 
    {
      memcpy((buf->buf)+total_len, linemembuffer, len);
      total_len=total_len+len;
    }
  }
  
  buf->size = total_len;
  return total_len;
  
}
// read one line of intel hex from file. Return the number of databytes
// Since the arduino code is always sequential, ignore the address for now.
// If you want to burn bootloaders, etc. we'll have to modify to 
// return the address.

// INTEL HEX FORMAT:
// :<8-bit record size><16bit address><8bit record type><data...><8bit checksum>
int readIntelHexLine(File input, int *address, unsigned char *buf){
  unsigned char c;
  int i=0;
  while (true){
    if (input.available()){
      c = input.read();
      // this should handle unix or ms-dos line endings.
      // break out when you reach either, then check
      // for lf in stream to discard
      if ((c == 0x0d)|| (c == 0x0a))
        break;
      else
        linebuffer[i++] =c;
    }
    else return -1; //end of file
  }
  linebuffer[i]= 0; // terminate the string
  //peek at the next byte and discard if line ending char.
  if (input.peek() == 0xa)
    input.read();
  int len = hex2byte(&linebuffer[1]);
  *address = (hex2byte(&linebuffer[3]) <<8) |
               (hex2byte(&linebuffer[5]));
  int j=0;
  for (int i = 9; i < ((len*2)+9); i +=2){
    buf[j] = hex2byte(&linebuffer[i]);
    j++;
  }
  return len;
}
unsigned char hex2byte(unsigned char *code){
  unsigned char result =0;

  if ((code[0] >= '0') && (code[0] <='9')){
    result = ((int)code[0] - '0') << 4;
  }
  else if ((code[0] >='A') && (code[0] <= 'F')) {
    result = ((int)code[0] - 'A'+10) << 4;
  }
  if ((code[1] >= '0') && (code[1] <='9')){
    result |= ((int)code[1] - '0');
  }
  else if ((code[1] >='A') && (code[1] <= 'F'))  
    result |= ((int)code[1] -'A'+10);  
return result;
}

// Right now there is only one file.
void programArduino(char * filename){

  digitalWrite(rstPin,HIGH);

  unsigned int major=0;
  unsigned int minor=0;
  delay(100);
   toggle_Reset();
   delay(10);
   stk500_getsync();
   stk500_getparm(Parm_STK_SW_MAJOR, &major);
//  DEBUGP("software major: ");
//  DEBUGPLN(major);
  stk500_getparm(Parm_STK_SW_MINOR, &minor);
//  DEBUGP("software Minor: ");
//  DEBUGPLN(minor);
if (SD.exists(filename)){
    myFile = SD.open(filename, FILE_READ);
  }
  else{
//    DEBUGP(filename);
//    DEBUGPLN(" doesn't exist");
    return;
  }
  //enter program mode
  stk500_program_enable();


  while (readPage(myFile,&mybuf) > 0){
    stk500_loadaddr(mybuf.pageaddress>>1);
    stk500_paged_write(&mybuf, mybuf.size, mybuf.size);
  }

  // could verify programming by reading back pages and comparing but for now, close out
  stk500_disable();
  delay(10);
  toggle_Reset();
  myFile.close();
  blinky(4,500);
  
  
}
void blinky(int times, long delaytime){
  for (int i = 0 ; i < times; i++){
    //digitalWrite(LED1,HIGH);
    delay(delaytime);
   // digitalWrite(LED1, LOW);
    delay (delaytime);
  }
  
}
void toggle_Reset()
{
  digitalWrite(rstPin, LOW);
  delayMicroseconds(1000);
  digitalWrite(rstPin,HIGH);
}
static int stk500_send(byte *buf, unsigned int len)
{

  Serial.write(buf,len);
}
static int stk500_recv(byte * buf, unsigned int len)
{
  int rv;

 
  rv = Serial.readBytes((char *)buf,len);
  if (rv < 0) {
    return -1;
  }
  return 0;
}
int stk500_drain()
{
  while (Serial.available()> 0)
  {  
//    DEBUGP("draining: ");
//    DEBUGPLN(Serial.read(),HEX);
    Serial.read();
  }
  return 1;
}
int stk500_getsync()
{
  byte buf[32], resp[32];

  /*
   * get in sync */
  buf[0] = Cmnd_STK_GET_SYNC;
  buf[1] = Sync_CRC_EOP;
  
  /*
   * First send and drain a few times to get rid of line noise 
   */
  
  stk500_send(buf, 2);
  stk500_drain();
  stk500_send(buf, 2);
  stk500_drain();
  
  stk500_send(buf, 2);
  if (stk500_recv(resp, 1) < 0)
    return -1;
  if (resp[0] != Resp_STK_INSYNC) {
    stk500_drain();
    return -1;
  }

  if (stk500_recv(resp, 1) < 0)
    return -1;
  if (resp[0] != Resp_STK_OK) {
    return -1;
  }
  return 0;
}
static int stk500_getparm(unsigned parm, unsigned * value)
{
  byte buf[16];
  unsigned v;
  int tries = 0;

 retry:
  tries++;
  buf[0] = Cmnd_STK_GET_PARAMETER;
  buf[1] = parm;
  buf[2] = Sync_CRC_EOP;

  stk500_send(buf, 3);

  if (stk500_recv(buf, 1) < 0)
    exit(1);
  if (buf[0] == Resp_STK_NOSYNC) {
    if (tries > 33) {
      return -1;
    }
   if (stk500_getsync() < 0)
      return -1;
      
    goto retry;
  }
  else if (buf[0] != Resp_STK_INSYNC) {
    return -2;
  }

  if (stk500_recv(buf, 1) < 0)
    exit(1);
  v = buf[0];

  if (stk500_recv(buf, 1) < 0)
    exit(1);
  if (buf[0] == Resp_STK_FAILED) {
    return -3;
  }
  else if (buf[0] != Resp_STK_OK) {
    return -3;
  }

  *value = v;

  return 0;
}
/* read signature bytes - arduino version */
static int arduino_read_sig_bytes(AVRMEM * m)
{
  unsigned char buf[32];

  /* Signature byte reads are always 3 bytes. */

  if (m->size < 3) {
//    DEBUGPLN("memsize too small for sig byte read");
    return -1;
  }

  buf[0] = Cmnd_STK_READ_SIGN;
  buf[1] = Sync_CRC_EOP;

  stk500_send(buf, 2);

  if (stk500_recv(buf, 5) < 0)
    return -1;
  if (buf[0] == Resp_STK_NOSYNC) {
	return -1;
  } else if (buf[0] != Resp_STK_INSYNC) {
	return -2;
  }
  if (buf[4] != Resp_STK_OK) {
    return -3;
  }

  m->buf[0] = buf[1];
  m->buf[1] = buf[2];
  m->buf[2] = buf[3];

  return 3;
}

static int stk500_loadaddr(unsigned int addr)
{
  unsigned char buf[16];
  int tries;

  tries = 0;
 retry:
  tries++;
  buf[0] = Cmnd_STK_LOAD_ADDRESS;
  buf[1] = addr & 0xff;
  buf[2] = (addr >> 8) & 0xff;
  buf[3] = Sync_CRC_EOP;


  stk500_send(buf, 4);

  if (stk500_recv(buf, 1) < 0)
    exit(1);
  if (buf[0] == Resp_STK_NOSYNC) {
    if (tries > 33) {
      return -1;
    }
    if (stk500_getsync() < 0)
      return -1;
    goto retry;
  }
  else if (buf[0] != Resp_STK_INSYNC) {
    return -1;
  }

  if (stk500_recv(buf, 1) < 0)
    exit(1);
  if (buf[0] == Resp_STK_OK) {
    return 0;
  }

  return -1;
}
static int stk500_paged_write(AVRMEM * m, 
                              int page_size, int n_bytes)
{
  // This code from avrdude has the luxury of living on a PC and copying buffers around.
  // not for us...
 // unsigned char buf[page_size + 16];
 unsigned char cmd_buf[16]; //just the header
  int memtype;
 // unsigned int addr;
  int block_size;
  int tries;
  unsigned int n;
  unsigned int i;
  int flash;

  // Fix page size to 128 because that's what arduino expects
  page_size = 128;
  //EEPROM isn't supported
  memtype = 'F';
  flash = 1;


    /* build command block and send data separeately on arduino*/
    
    i = 0;
    cmd_buf[i++] = Cmnd_STK_PROG_PAGE;
    cmd_buf[i++] = (page_size >> 8) & 0xff;
    cmd_buf[i++] = page_size & 0xff;
    cmd_buf[i++] = memtype;
    stk500_send(cmd_buf,4);
    stk500_send(&m->buf[0], page_size);
    cmd_buf[0] = Sync_CRC_EOP;
    stk500_send( cmd_buf, 1);

    if (stk500_recv(cmd_buf, 1) < 0)
      exit(1); // errr need to fix this... 
    if (cmd_buf[0] == Resp_STK_NOSYNC) {
        return -3;
     }
    else if (cmd_buf[0] != Resp_STK_INSYNC) {

      return -4;
    }
    
    if (stk500_recv(cmd_buf, 1) < 0)
      exit(1);
    if (cmd_buf[0] != Resp_STK_OK) {

      return -5;
    }
  

  return n_bytes;
}
#ifdef LOADVERIFY //maybe sometime? note code needs to be re-written won't work as is
static int stk500_paged_load(AVRMEM * m, 
                             int page_size, int n_bytes)
{
  unsigned char buf[16];
  int memtype;
  unsigned int addr;
  int a_div;
  int tries;
  unsigned int n;
  int block_size;

  memtype = 'F';


  a_div = 1;

  if (n_bytes > m->size) {
    n_bytes = m->size;
    n = m->size;
  }
  else {
    if ((n_bytes % page_size) != 0) {
      n = n_bytes + page_size - (n_bytes % page_size);
    }
    else {
      n = n_bytes;
    }
  }

  for (addr = 0; addr < n; addr += page_size) {
//    report_progress (addr, n_bytes, NULL);

    if ((addr + page_size > n_bytes)) {
	   block_size = n_bytes % page_size;
	}
	else {
	   block_size = page_size;
	}
  
    tries = 0;
  retry:
    tries++;
    stk500_loadaddr(addr/a_div);
    buf[0] = Cmnd_STK_READ_PAGE;
    buf[1] = (block_size >> 8) & 0xff;
    buf[2] = block_size & 0xff;
    buf[3] = memtype;
    buf[4] = Sync_CRC_EOP;
    stk500_send(buf, 5);

    if (stk500_recv(buf, 1) < 0)
      exit(1);
    if (buf[0] == Resp_STK_NOSYNC) {
      if (tries > 33) {
        error(ERRORNOSYNC);
        return -3;
      }
      if (stk500_getsync() < 0)
	return -1;
      goto retry;
    }
    else if (buf[0] != Resp_STK_INSYNC) {
      error1(ERRORPROTOSYNC, buf[0]);
      return -4;
    }

    if (stk500_recv(&m->buf[addr], block_size) < 0)
      exit(1);

    if (stk500_recv(buf, 1) < 0)
      exit(1);

    if (buf[0] != Resp_STK_OK) {
        error1(ERRORPROTOSYNC, buf[0]);
        return -5;
      }
    }
  

  return n_bytes;
}
#endif

/*
 * issue the 'program enable' command to the AVR device
 */
static int stk500_program_enable()
{
  unsigned char buf[16];
  int tries=0;

 retry:
  
  tries++;

  buf[0] = Cmnd_STK_ENTER_PROGMODE;
  buf[1] = Sync_CRC_EOP;

  stk500_send( buf, 2);
  if (stk500_recv( buf, 1) < 0)
    exit(1);
  if (buf[0] == Resp_STK_NOSYNC) {
    if (tries > 33) {
      return -1;
    }
    if (stk500_getsync()< 0)
      return -1;
    goto retry;
  }
  else if (buf[0] != Resp_STK_INSYNC) {
    return -1;
  }

  if (stk500_recv( buf, 1) < 0)
    exit(1);
  if (buf[0] == Resp_STK_OK) {
    return 0;
  }
  else if (buf[0] == Resp_STK_NODEVICE) {
    return -1;
  }

  if(buf[0] == Resp_STK_FAILED)
  {
	  return -1;
  }



  return -1;
}

static void stk500_disable()
{
  unsigned char buf[16];
  int tries=0;

 retry:
  
  tries++;

  buf[0] = Cmnd_STK_LEAVE_PROGMODE;
  buf[1] = Sync_CRC_EOP;

  stk500_send( buf, 2);
  if (stk500_recv( buf, 1) < 0)
    exit(1);
  if (buf[0] == Resp_STK_NOSYNC) {
    if (tries > 33) {
      return;
    }
    if (stk500_getsync() < 0)
      return;
    goto retry;
  }
  else if (buf[0] != Resp_STK_INSYNC) {
    return;
  }

  if (stk500_recv( buf, 1) < 0)
    exit(1);
  if (buf[0] == Resp_STK_OK) {
    return;
  }
  else if (buf[0] == Resp_STK_NODEVICE) {
    return;
  }


  return;
}


